/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_EXPECTATION_HPP
#define QT_MOCKS_EXPECTATION_HPP

#include <QtCore>

#include <functional>
#include <memory>
#include <optional>
#include <queue>
#include <tuple>
#include <type_traits>
#include <unordered_map>

#include <qt-mocks/qt-mocks_export.h>

#include "Mock.hpp"

class QString;

/**
 * @brief Specifies and verifies the calls to a member function on a mocked object. This sets up an
 * expectation for a call to a member function on a mocked object. It allows you to configure
 * specific behaviors and conditions for the mock function when it is invoked in your tests.
 *
 * @param expectations The object, which holds the mocked object and the required function
 * expectations.
 * @param memberFunctionName The name of the member function that is expected to be called.
 * @param ParameterTypes The types of the parameters that the member function accepts.
 * @param specifiers Additional specifiers for the member function, such as 'const'.
 *
 * @details
 * This macro encapsulates the process of setting up a mocked member function call for the given
 * expectation object.
 * The macro performs the following tasks:
 *   - Sets the function pointer of the mocked member function to a lambda that processes the call
 *   based on specified expectations.
 *   - Initializes the arguments for the mocked member function.
 *   - Creates and pushes a unique expectation object for the specified member function.
 *
 * @note To use this macro:
 * - Ensure that the `expectations` object is of the `Expectations` class which wraps a mock object.
 * - The macro prepares the mock for the specified member function, making it immediately ready to
 * receive expectations.
 *
 * Example Usage:
 * @code{.cpp}
 * // Create the mock object and the expectations object
 * MockClass mock;
 * qt_mocks::Expectations<MockClass> expectations{mock};
 *
 * // Use the macro to setup expectations for the member function 'foo' with specific parameter
 * types and no specifiers
 * QVERIFY_CALL(expectations, foo, (int, double), ());
 *		.will(qt_mocks::give(7)); // Configure the expectation (can be configured further)
 * @endcode
 *
 * The above code sets up the mock object so that when the member function `foo(int, double)` is
 * called, the mocked function will use the arguments `42` and `3.14`, it will return `7`.
 */
#define QVERIFY_CALL(expectations, memberFunctionName, ParameterTypes, specifiers)  \
		QMOCK_BIND(expectations.mockedObject_, memberFunctionName, ParameterTypes, specifiers) =  \
				[&expectations] (auto ... args) -> decltype(expectations.mockedObject_._QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers).gives()) \
				{  \
					using ReturnType = decltype(expectations.mockedObject_._QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers).gives());  \
					using ExpectationType = qt_mocks::internals::MockedExpectation<ReturnType, _QT_MOCKS_SPLIT_PARAMETERS ParameterTypes>;  \
					auto& queue{expectations.functionExpectationsQueue_[_QT_MOCKS_STRINGIFY_(_QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers))]}; \
					return qt_mocks::internals::ExpectationQueueProcessor<ExpectationType, ReturnType>::processNext(queue, __FILE__, __LINE__, args...);  \
				};  \
		expectations.mockedObject_._QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers).arguments_ = {};  \
		qt_mocks::internals::ExpectationBuilder<decltype(expectations.mockedObject_._QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers).gives()), _QT_MOCKS_SPLIT_PARAMETERS ParameterTypes>{  \
			expectations.functionExpectationsQueue_[_QT_MOCKS_STRINGIFY_(_QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers))].push(  \
				std::make_unique<qt_mocks::internals::MockedExpectation<decltype(expectations.mockedObject_._QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers).gives()), _QT_MOCKS_SPLIT_PARAMETERS ParameterTypes>>(  \
					#memberFunctionName,  \
					__FILE__,  \
					__LINE__  \
				)  \
			)  \
		}

namespace qt_mocks::internals
{
class ExpectationQueue;
}

namespace qt_mocks
{
/**
 * @brief A manager for collecting all function expectations for a given mocked object. Each
 * expectation will be expected to be invoked in the order in which the expectations are built.
 *
 * @tparam MockedObjectType The type of the object that is mocked.
 */
template<typename MockedObjectType>
struct Expectations
{
	/**
	 * Full initialisation constructor.
	 *
	 * @param mockedObject A reference to the mocked object to which expectations will be applied.
	 */
	explicit Expectations(MockedObjectType& mockedObject);

	/**
	 * The mocked object being tested. This member holds the actual instance of the mocked object to
	 * which expectations are applied. It serves as the primary interface through which the mock
	 * framework interacts with the object under test.
	 */
	MockedObjectType& mockedObject_;
	/**
	 * Stores the expectations for functions in the mock framework. This holds the expectations for
	 * various mocked function calls. The key is the mocked function name ID as a string, and the
	 * value is a container of the sequence of expectations for that function. This allows the
	 * mocking system to keep track of what actions are expected when specific functions are invoked
	 * on a mock object.
	 */
	std::unordered_map<std::string, internals::ExpectationQueue> functionExpectationsQueue_;
};
/**
 * @brief Encapsulates a value to be returned by a mocked function.
 *
 * @tparam ReturnType The type of the value to be returned.
 */
template<typename ReturnType>
class ValueGiver final
{
	ReturnType value_{};

public:
	/**
	 * @brief Full initialisation constructor that species the value to return for a mocked function.
	 *
	 * @param value The initial value to be given.
	 */
	explicit ValueGiver(ReturnType value);
	/**
	 * Provides the stored value for a mocked function.
	 *
	 * @return The stored value.
	 */
	ReturnType giveValue() const noexcept;
};
template<typename ExceptionType>
/**
 * @brief Encapsulates an error to be raised by a mocked function.
 *
 * @tparam ExceptionType The type of the exception to be raised.
 */
class ErrorRaiser final
{
	ExceptionType error_;

public:
	/**
	 * @brief Full initialization constructor.
	 *
	 * @tparam ExceptionArguments The argument types used to construct the error.
	 *
	 * @param arguments A list of arguments used to construct the error object.
	 */
	template<typename ... ExceptionArguments>
	explicit ErrorRaiser(ExceptionArguments&& ... arguments);
	/**
	 * @brief This throws the exception that was stored.
	 *
	 * @throws ExceptionType The stored exception to be thrown.
	 */
	void doThrow() const;
};
/**
 * @brief Encapsulates a value to be generated by a mocked function.
 *
 * @tparam ReturnType The type of the value to be generated.
 */
template<typename ReturnType>
class ValueGenerator final
{
	std::function<ReturnType()> generator_{};

public:
	/**
	 * @brief Full initialisation constructor that initialises the functor that generates a value
	 * for a mocked function.
	 *
	 * @param generator The functor that creates the desired return value upon each invocation.
	 */
	explicit ValueGenerator(std::function<ReturnType()> generator);
	/**
	 * Invokes the value generator and returns a generated value.
	 *
	 * @return A generated value of the specified ReturnType.
	 */
	ReturnType operator()() const;
};

/**
 * Creates a means to specify exactly what value should a mocked function return.
 *
 * @tparam ReturnType The type of the value to be returned.
 *
 * @param value The value that should be returned.
 *
 * @return an object containing all the information on what the mocked function should return.
 */
template<typename ReturnType>
ValueGiver<ReturnType> give(ReturnType value) noexcept;
/**
 * Creates a means to specify exactly what error exception should a mocked function throw.
 *
 * @tparam ExceptionType The type of the exception to raise.
 * @tparam ExceptionArguments The types of the parameters for the exception being raised.
 *
 * @param arguments Any arguments to supply to the exception being raised.
 *
 * @return an object containing all the information on what the mocked function should throw.
 */
template<typename ExceptionType, typename ... ExceptionArguments>
ErrorRaiser<ExceptionType> raise(ExceptionArguments&& ... arguments);
/**
 * Creates a means to generate a value for a mocked function's return value.
 *
 * @tparam FunctionType The callable type that takes no arguments and returns the value expected by
 * the mocked function.
 *
 * @param generator The functor that generates a value upon each invocation.
 *
 * @return an object containing all the information on what the mocked function should generate.
 */
template<typename FunctionType>
auto generate(FunctionType&& generator);

}

namespace qt_mocks::internals
{
enum class QT_MOCKS_EXPORT ExpectedAction
{
	Invoked,
	GiveValue,
	RaiseError,
	GenerateValue
};

class QT_MOCKS_EXPORT Expectation
{
	const char* functionName_;
	const char* fileName_;
	int lineNumber_;
	bool isRequired_;
	std::optional<::quint64> expectedInvokeCount_;
	::quint64 actualInvokedCount_;

public:
	Expectation(
		const char* const functionName,
		const char* const fileName,
		const int lineNumber
	);
	Expectation(const Expectation&) = delete;
	Expectation(Expectation&&) = default;
	Expectation& operator=(const Expectation&) = delete;
	Expectation& operator=(Expectation&&) = default;
	virtual ~Expectation();

	[[nodiscard]]
	const char* fileName() const noexcept;

	[[nodiscard]]
	int lineNumber() const noexcept;

	[[nodiscard]]
	const char* functionName() const noexcept;

	[[nodiscard]]
	::QString description() const noexcept;

	[[nodiscard]]
	bool isInvokeRequired() const noexcept;

	[[nodiscard]]
	const std::optional<::quint64>& expectedInvokeCount() const noexcept;

	[[nodiscard]]
	::quint64 actualInvokedCount() const noexcept;

	void markInvokeRequired() noexcept;

	void setExpectedInvokeCount(const ::quint64 count) noexcept;

	void increaseInvokeCount() noexcept;

	ExpectedAction action_;
};

template<typename ReturnType, typename ... ParameterTypes>
class MockedExpectation final : public Expectation
{
public:
	MockedExpectation(
		const char* const functionName,
		const char* const fileName,
		const int lineNumber
	);

	Mock<ReturnType, ParameterTypes ...> mock_{};
	std::function<void(
		std::tuple<std::remove_const_t<std::remove_reference_t<ParameterTypes>> ...>
	)> argumentVerifier_{nullptr};
};

class QT_MOCKS_EXPORT ExpectationQueue final
{
	std::queue<std::unique_ptr<Expectation>> expectations_;

public:
	ExpectationQueue() = default;
	ExpectationQueue(const ExpectationQueue&) = delete;
	ExpectationQueue(ExpectationQueue&&) = default;
	ExpectationQueue& operator=(const ExpectationQueue&) = delete;
	ExpectationQueue& operator=(ExpectationQueue&&) = default;
	~ExpectationQueue();

	const std::unique_ptr<Expectation>& peek() const noexcept;

	Expectation& push(std::unique_ptr<Expectation> expectation);

	void pop();
};
/**
 * @brief This is used to build and manage the expectations for mocked objects.
 */
template<typename ReturnType, typename ... ParameterTypes>
class ExpectationBuilder final
{
	MockedExpectation<ReturnType, ParameterTypes ...>& expectation_;

public:
	/**
	 * Full initialisation constructor.
	 *
	 * @param expectation The expectation object whose expected behaviour will be built.
	 */
	explicit ExpectationBuilder(Expectation& expectation);
	/**
	 * Sets the expected return value for a mocked function.
	 *
	 * @tparam CheckReturnType The type of the return value.
	 *
	 * @param valueGiver Object encapsulating the value to be returned by the mocked function.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	template<typename CheckReturnType = ReturnType>
	std::enable_if_t<! std::is_void_v<CheckReturnType>, ExpectationBuilder&> will(
		ValueGiver<ReturnType> valueGiver
	);
	/**
	 * Sets the expectation to raise an exception when the mocked function is called.
	 *
	 * @tparam ExceptionType The expected type of exception to be raised.
	 *
	 * @param errorRaiser An instance that specifies the exception type and data to raise.
	 * 
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	template<typename ExceptionType>
	ExpectationBuilder& will(ErrorRaiser<ExceptionType> errorRaiser);
	/**
	 * Sets the expected return value generator for a mocked function.
	 *
	 * @tparam CheckReturnType The type of the return value.
	 *
	 * @param valueGenerator Object encapsulating the generator used to create the value to be
	 * returned by the mocked function.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	template<typename CheckReturnType = ReturnType>
	std::enable_if_t<! std::is_void_v<CheckReturnType>, ExpectationBuilder&> will(
		ValueGenerator<ReturnType> valueGenerator
	);
	/**
	 * Sets the expected number of times the mocked function should be called.
	 *
	 * @param count The number of times the function is expected to be called.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& times(const ::quint64 count);
	/**
	 * Marks the expectation as required, meaning that it must be invoked.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& require();
	/**
	 * Specifies the matchers to be used for the parameters of the expectation.
	 *
	 * @bug This is not fully implemented yet.
	 *
	 * @tparam MatcherTypes The type for each matcher. Each matcher type must take the same type as
	 * the mocked function in the same order as found in the mocked function parameters.
	 *
	 * @param matchers Matcher objects to compare the arguments against.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	template<typename ... MatcherTypes>
	ExpectationBuilder& arguments(MatcherTypes ... matchers);
};

template<typename ReturnType>
class ExpectationBuilder<ReturnType, void> final
{
	MockedExpectation<ReturnType, void>& expectation_;

public:
	/**
	 * Full initialisation constructor.
	 *
	 * @param expectation The expectation object whose expected behaviour will be built.
	 */
	explicit ExpectationBuilder(Expectation& expectation);
	/**
	 * Sets the expected return value for a mocked function.
	 *
	 * @param valueGiver Object encapsulating the value to be returned by the mocked function.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& will(ValueGiver<ReturnType> valueGiver);
	/**
	 * Sets the expectation to raise an exception when the mocked function is called.
	 *
	 * @tparam ExceptionType The expected type of exception to be raised.
	 *
	 * @param errorRaiser An instance that specifies the exception type and data to raise.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	template<typename ExceptionType>
	ExpectationBuilder& will(ErrorRaiser<ExceptionType> errorRaiser);
	/**
	 * Sets the expected return value generator for a mocked function.
	 *
	 * @param valueGenerator Object encapsulating the generator used to create the value to be
	 * returned by the mocked function.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& will(ValueGenerator<ReturnType> valueGenerator);
	/**
	 * Sets the expected number of times the mocked function should be called.
	 *
	 * @param count The number of times the function is expected to be called.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& times(const ::quint64 count);
	/**
	 * Marks the expectation as required, meaning that it must be invoked.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& require();
};

template<>
class QT_MOCKS_EXPORT ExpectationBuilder<void, void> final
{
	MockedExpectation<void, void>& expectation_;

public:
	/**
	 * Full initialisation constructor.
	 *
	 * @param expectation The expectation object whose expected behaviour will be built.
	 */
	explicit ExpectationBuilder(Expectation& expectation);
	/**
	 * Sets the expectation to raise an exception when the mocked function is called.
	 *
	 * @tparam ExceptionType The expected type of exception to be raised.
	 *
	 * @param errorRaiser An instance that specifies the exception type and data to raise.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	template<typename ExceptionType>
	ExpectationBuilder& will(ErrorRaiser<ExceptionType> errorRaiser);
	/**
	 * Sets the expected number of times the mocked function should be called.
	 *
	 * @param count The number of times the function is expected to be called.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& times(const ::quint64 count);
	/**
	 * Marks the expectation as required, meaning that it must be invoked.
	 *
	 * @return A reference to the updated builder object, allowing for method chaining.
	 */
	ExpectationBuilder& require();
};

template<typename ExpectationType, typename ReturnType>
struct ExpectationQueueProcessor final
{
	ExpectationQueueProcessor() = delete;

	template<typename QueueType, typename ... Args>
	static ReturnType processNext(
		QueueType& queue,
		const char* const fileName,
		const int lineNumber,
		Args&& ... args
	);
};

template<typename ExpectationType>
struct ExpectationQueueProcessor<ExpectationType, void> final
{
	ExpectationQueueProcessor() = delete;

	template<typename QueueType, typename ... Args>
	static void processNext(
		QueueType& queue,
		const char* const fileName,
		const int lineNumber,
		Args&& ... args
	);
};

template<typename ReturnType, typename ExpectationType, typename ... ParameterTypes>
ReturnType process(
	ExpectationQueue& queue,
	ExpectationType& expectation,
	ParameterTypes ... passedInArgs
);

QT_MOCKS_EXPORT ::QString toStdString(const ExpectedAction action);
}

#define _QT_MOCKS_STRINGIFY_IMPLEMENTATION_(x) #x
#define _QT_MOCKS_STRINGIFY_(x) _QT_MOCKS_STRINGIFY_IMPLEMENTATION_(x)

#include "Expectation.tpp"

#endif
