/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_EXPECTATION_HPP
#error Should not be included directly.
#endif

#include <QTest>

template<typename ReturnType, typename ... ParameterTypes>
qt_mocks::internals::MockedExpectation<ReturnType, ParameterTypes ...>::MockedExpectation(
	const char* const functionName,
	const char* const fileName,
	const int lineNumber
) :
	Expectation{functionName, fileName, lineNumber}
{}

template<typename MockedObjectType>
qt_mocks::Expectations<MockedObjectType>::Expectations(MockedObjectType& mockedObject) :
	mockedObject_{mockedObject},
	functionExpectationsQueue_{}
{}

template<typename ReturnType>
qt_mocks::ValueGiver<ReturnType> qt_mocks::give(ReturnType value) noexcept
{
	if constexpr (std::is_reference_v<ReturnType>)
		return ValueGiver<ReturnType>{value};
	else
		return ValueGiver<ReturnType>{std::move(value)};
}

template<typename ReturnType>
qt_mocks::ValueGiver<ReturnType>::ValueGiver(ReturnType value) :
	value_{value}
{}

template<typename ReturnType>
ReturnType qt_mocks::ValueGiver<ReturnType>::giveValue() const noexcept
{
	return this->value_;
}

template<typename ExceptionType, typename ... ExceptionArguments>
qt_mocks::ErrorRaiser<ExceptionType> qt_mocks::raise(ExceptionArguments&& ... arguments)
{
	return ErrorRaiser<ExceptionType>{std::forward<ExceptionArguments>(arguments) ...};
}

template<typename FunctionType>
auto qt_mocks::generate(FunctionType&& generator)
{
	using ReturnType = std::invoke_result_t<FunctionType>;
	return ValueGenerator<ReturnType>{
		std::function<ReturnType()>{std::forward<FunctionType>(generator)}
	};
}

template<typename ExceptionType>
template<typename ... ExceptionArguments>
qt_mocks::ErrorRaiser<ExceptionType>::ErrorRaiser(ExceptionArguments&& ... arguments) :
	error_{std::forward<ExceptionArguments>(arguments) ...}
{}

template<typename ExceptionType>
void qt_mocks::ErrorRaiser<ExceptionType>::doThrow() const
{
	throw this->error_;
}

template<typename ReturnType>
qt_mocks::ValueGenerator<ReturnType>::ValueGenerator(std::function<ReturnType()> generator) :
	generator_{std::move(generator)}
{
	assert(this->generator_ != nullptr);
}

template<typename ReturnType>
ReturnType qt_mocks::ValueGenerator<ReturnType>::operator()() const
{
	return this->generator_();
}

//{ Builder With Arguments

template<typename ReturnType, typename ... ParameterTypes>
qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>::ExpectationBuilder(
	Expectation& expectation
) :
	expectation_{dynamic_cast<MockedExpectation<ReturnType, ParameterTypes ...>&>(expectation)}
{}

template<typename ReturnType, typename ... ParameterTypes>
template<typename CheckReturnType>
std::enable_if_t<
	 ! std::is_void_v<CheckReturnType>,
	qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>&
> qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>::will(
			ValueGiver<ReturnType> valueGiver
		)
{
	this->expectation_.action_ = ExpectedAction::GiveValue;
	this->expectation_.mock_.function_ =
			[giver{std::move(valueGiver)}] (ParameterTypes ...) -> ReturnType
			{
				return giver.giveValue();
			};
	return *this;
}

template<typename ReturnType, typename ... ParameterTypes>
template<typename ExceptionType>
qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>::will(
			ErrorRaiser<ExceptionType> errorRaiser
		)
{
	this->expectation_.action_ = ExpectedAction::RaiseError;
	this->expectation_.mock_.function_ =
			[raiser{std::move(errorRaiser)}] (ParameterTypes ...) -> ReturnType
			{
				raiser.doThrow();

				if constexpr (std::is_reference_v<ReturnType>)
				{
					static std::remove_reference_t<ReturnType> dummy{};
					return dummy;
				}
				else
					return {};
			};
	return *this;
}

template<typename ReturnType, typename ... ParameterTypes>
template<typename CheckReturnType>
std::enable_if_t<
	 ! std::is_void_v<CheckReturnType>,
	qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes...>&
> qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes...>::will(
	ValueGenerator<ReturnType> valueGenerator
)
{
	this->expectation_.action_ = ExpectedAction::GenerateValue;
	this->expectation_.mock_.function_ =
			[generator{std::move(valueGenerator)}] (ParameterTypes ...) -> ReturnType
			{
				return generator();
			};
	return *this;
}

template<typename ReturnType, typename ... ParameterTypes>
qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>::times(
			const ::quint64 count
		)
{
	this->expectation_.setExpectedInvokeCount(count);
	return *this;
}

template<typename ReturnType, typename ... ParameterTypes>
qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>::require()
{
	this->expectation_.markInvokeRequired();
	return *this;
}

template<typename ReturnType, typename ... ParameterTypes>
template<typename ... MatcherTypes>
qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, ParameterTypes ...>::arguments(
			MatcherTypes ... matchers
		)
{
	this->expectation_.argumentVerifier_ =
			[=, &expectation{this->expectation_}] (
				std::tuple<
					std::remove_const_t<std::remove_reference_t<ParameterTypes>> ...
				> arguments
			)
			{
				std::apply(
					[&] (const auto& ... args)
					{
						(
							(
								QTest::qVerify(
									matchers.compare(args),
									_QT_MOCKS_STRINGIFY_(args),
									matchers.description(args).c_str(),
									expectation.fileName(),
									expectation.lineNumber()
								)
							), ...
						);
					},
					arguments
				);
			};
	return *this;
}

//}

//{ Builder Without Arguments

template<typename ReturnType>
qt_mocks::internals::ExpectationBuilder<ReturnType, void>::ExpectationBuilder(
	Expectation& expectation
) :
	expectation_{dynamic_cast<MockedExpectation<ReturnType, void>&>(expectation)}
{}

template<typename ReturnType>
qt_mocks::internals::ExpectationBuilder<ReturnType, void>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, void>::will(
			ValueGiver<ReturnType> valueGiver
		)
{
	this->expectation_.action_ = ExpectedAction::GiveValue;
	this->expectation_.mock_.function_ =
			[giver{std::move(valueGiver)}]() -> ReturnType
			{
				return giver.giveValue();
			};
	return *this;
}

template<typename ReturnType>
template<typename ExceptionType>
qt_mocks::internals::ExpectationBuilder<ReturnType, void>&
		qt_mocks::internals::ExpectationBuilder< ReturnType, void>::will(
			ErrorRaiser<ExceptionType> errorRaiser
		)
{
	this->expectation_.action_ = ExpectedAction::RaiseError;
	this->expectation_.mock_.function_ =
			[raiser{std::move(errorRaiser)}]() -> ReturnType
			{
				raiser.doThrow();

				if constexpr (std::is_reference_v<ReturnType>)
				{
					static std::remove_reference_t<ReturnType> dummy{};
					return dummy;
				}
				else
					return {};
			};
	return *this;
}

template<typename ReturnType>
qt_mocks::internals::ExpectationBuilder<ReturnType, void>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, void>::will(
			ValueGenerator<ReturnType> valueGenerator
		)
{
	this->expectation_.action_ = ExpectedAction::GenerateValue;
	this->expectation_.mock_.function_ =
			[generator{std::move(valueGenerator)}] () -> ReturnType
			{
				return generator();
			};
	return *this;
}

template<typename ReturnType>
qt_mocks::internals::ExpectationBuilder<ReturnType, void>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, void>::times(const ::quint64 count)
{
	this->expectation_.setExpectedInvokeCount(count);
	return *this;
}

template<typename ReturnType>
qt_mocks::internals::ExpectationBuilder<ReturnType, void>&
		qt_mocks::internals::ExpectationBuilder<ReturnType, void>::require()
{
	this->expectation_.markInvokeRequired();
	return *this;
}

template<typename ExceptionType>
qt_mocks::internals::ExpectationBuilder<void, void>&
		qt_mocks::internals::ExpectationBuilder<void, void>::will(
			ErrorRaiser<ExceptionType> errorRaiser
		)
{
	this->expectation_.action_ = ExpectedAction::RaiseError;
	this->expectation_.mock_.function_ =
			[raiser{std::move(errorRaiser)}]()
			{
				raiser.doThrow();
			};
	return *this;
}

//}

//{ ExpectationQueueProcessor

template<typename ExpectationType, typename ReturnType>
template<typename QueueType, typename ... Args>
ReturnType qt_mocks::internals::ExpectationQueueProcessor<ExpectationType, ReturnType>::processNext(
	QueueType& queue,
	const char* const fileName,
	const int lineNumber,
	Args&& ... args
)
{
	auto& next{queue.peek()};

	if (next == nullptr)
	{
		QTest::qFail(
			"No additional expectation queued for the mocked function.",
			fileName,
			lineNumber
		);

		if constexpr (std::is_reference_v<ReturnType>)
		{
			static std::remove_reference_t<ReturnType> dummy{};
			return dummy;
		}
		else
			return ReturnType{};
	}
	auto& nextExpectation{dynamic_cast<ExpectationType&>(*next)};
	return internals::process<ReturnType>(queue, nextExpectation, std::forward<Args>(args) ...);
}

template<typename ExpectationType>
template<typename QueueType, typename ... Args>
void qt_mocks::internals::ExpectationQueueProcessor<ExpectationType, void>::processNext(
	QueueType& queue,
	const char* const fileName,
	const int lineNumber,
	Args&& ... args
)
{
	auto& next{queue.peek()};

	if (next == nullptr)
	{
		QTest::qFail(
			"No additional expectation queued for the mocked function.",
			fileName,
			lineNumber
		);
		return;
	}
	auto& nextExpectation{dynamic_cast<ExpectationType&>(*next)};
	internals::process<void>(queue, nextExpectation, std::forward<Args>(args) ...);
}

//}

template<typename ReturnType, typename ExpectationType, typename ... ParameterTypes>
ReturnType qt_mocks::internals::process(
	ExpectationQueue& queue,
	ExpectationType& expectation,
	ParameterTypes ... passedInArgs
)
{
	expectation.increaseInvokeCount();

	if constexpr (std::is_reference_v<ReturnType>)
	{
		auto& toReturn{internals::invokeImplementation(expectation.mock_)};

		if constexpr (std::tuple_size_v<decltype(expectation.mock_.arguments_)> > 0)
		{
			if (expectation.argumentVerifier_ != nullptr)
				expectation.argumentVerifier_(
					std::make_tuple(std::forward<ParameterTypes>(passedInArgs) ...)
				);
		}

		if (
			expectation.expectedInvokeCount().has_value()
			&& expectation.actualInvokedCount() == *expectation.expectedInvokeCount()
		)
			queue.pop();
		return toReturn;
	}
	else if constexpr (std::is_void_v<ReturnType>)
	{
		internals::invokeImplementation(expectation.mock_);

		if constexpr (std::tuple_size_v<decltype(expectation.mock_.arguments_)> > 0)
		{
			if (expectation.argumentVerifier_ != nullptr)
				expectation.argumentVerifier_(
					std::make_tuple(std::forward<ParameterTypes>(passedInArgs) ...)
				);
		}

		if (
			expectation.expectedInvokeCount().has_value()
			&& expectation.actualInvokedCount() == *expectation.expectedInvokeCount()
		)
			queue.pop();
		return;
	}
	else
	{
		auto toReturn{internals::invokeImplementation(expectation.mock_)};

		if constexpr (std::tuple_size_v<decltype(expectation.mock_.arguments_)> > 0)
		{
			if (expectation.argumentVerifier_ != nullptr)
				expectation.argumentVerifier_(
					std::make_tuple(std::forward<ParameterTypes>(passedInArgs) ...)
				);
		}

		if (
			expectation.expectedInvokeCount().has_value()
			&& expectation.actualInvokedCount() == *expectation.expectedInvokeCount()
		)
			queue.pop();
		return toReturn;
	}
}
