/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_MATCHERS_HPP_
#define QT_MOCKS_MATCHERS_HPP_

#include <QtGlobal>
#include <QTest>

#include <memory>
#include <ostream>
#include <sstream>
#include <string>
#include <tuple>
#include <typeinfo>
#include <utility>

#include <qt-mocks/qt-mocks_export.h>

#include "Output.hpp"
#include "Verify.hpp"


namespace qt_mocks::common_matchers
{

template<typename ValueType>
class ExpectedValue
{
	ValueType value_;

public:
	explicit ExpectedValue(ValueType value);
	ExpectedValue(const ExpectedValue&) = default;
	ExpectedValue(ExpectedValue&&) noexcept = default;
	~ExpectedValue() = default;
	ExpectedValue& operator=(const ExpectedValue&) = default;
	ExpectedValue& operator=(ExpectedValue&&) noexcept = default;
	const ValueType& value() const;

};
template<typename ValueType>
struct Matcher
{
	using MatcherValueType = ValueType;

	explicit Matcher(const ExpectedValue<ValueType>& value);
	explicit Matcher(std::unique_ptr<ExpectedValue<ValueType>> expected);
	Matcher(const Matcher&);
	Matcher(Matcher&& other) noexcept;
	virtual ~Matcher() = default;
	Matcher& operator=(const Matcher&) = delete;
	Matcher& operator=(Matcher&& other) noexcept;
	virtual bool compare(const ValueType& actual) const =0;
	virtual std::string description(const ValueType& actual) const =0;

	std::unique_ptr<ExpectedValue<ValueType>> expected_;
};
template<typename ValueType>
struct EqualMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<>
struct QT_MOCKS_EXPORT EqualMatcher<float>final : Matcher<float>
{
	using Matcher<float>::Matcher;

	bool compare(const float& actual) const override;
	std::string description(const float& actual) const override;
};
template<>
struct QT_MOCKS_EXPORT EqualMatcher<double>final : Matcher<double>
{
	using Matcher<double>::Matcher;

	bool compare(const double& actual) const override;
	std::string description(const double& actual) const override;
};
template<>
struct QT_MOCKS_EXPORT EqualMatcher<long double>final : Matcher<long double>
{
	using Matcher<long double>::Matcher;

	bool compare(const long double& actual) const override;
	std::string description(const long double& actual) const override;
};
template<typename ValueType>
struct NotEqualMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct LessThanMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct LessThanEqualMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct GreaterThanMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct GreaterThanEqualMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ClassType, typename MemberVariableType, typename MatcherType>
struct MemberVariableMatcher final : Matcher<ClassType>
{
	MemberVariableMatcher(MemberVariableType ClassType::* const member, const MatcherType& matcher);
	bool compare(const ClassType& actual) const override;
	std::string description(const ClassType& actual) const override;

	MemberVariableType ClassType::* const member_;
	MatcherType matcher_;
};
template<typename ClassType, typename MemberFunctionType, typename MatcherType>
struct MemberFunctionMatcher final : Matcher<ClassType>
{
	MemberFunctionMatcher(
		MemberFunctionType (ClassType::* member)() const,
		const MatcherType& matcher
	);
	bool compare(const ClassType& actual) const override;
	std::string description(const ClassType& actual) const override;

	MemberFunctionType (ClassType::* member_)() const;
	MatcherType matcher_;
};
template<typename ValueType>
struct BetweenExclusiveMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenExclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct BetweenBeginInclusiveMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenBeginInclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct BetweenEndInclusiveMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenEndInclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct BetweenInclusiveMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenInclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct NegationMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	template<typename MatcherType>
	explicit NegationMatcher(MatcherType&& matcher);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	std::unique_ptr<Matcher<ValueType>> matcher_;
};
struct QT_MOCKS_EXPORT BooleanMatcher final : Matcher<bool>
{
	explicit BooleanMatcher(const bool expected);
	bool compare(const bool& actual) const override;
	std::string description(const bool& actual) const override;
};
template<typename Container>
struct IsEmptyMatcher final : Matcher<Container>
{
	IsEmptyMatcher();
	bool compare(const Container& actual) const override;
	std::string description(const Container& actual) const override;
};
template<typename ValueType, typename ... MatcherTypes>
struct AllOfMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	explicit AllOfMatcher(MatcherTypes&& ... matchers);
	AllOfMatcher(const AllOfMatcher&) = delete;
	AllOfMatcher(AllOfMatcher&&) noexcept;
	~AllOfMatcher() override = default;
	AllOfMatcher& operator=(const AllOfMatcher&) = delete;
	AllOfMatcher& operator=(AllOfMatcher&&) noexcept;
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	std::tuple<MatcherTypes...> matchers_;
};
template<typename ValueType, typename ... MatcherTypes>
struct AnyOfMatcher final : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	explicit AnyOfMatcher(MatcherTypes&& ... matchers);
	AnyOfMatcher(const AnyOfMatcher&) = delete;
	AnyOfMatcher(AnyOfMatcher&&) noexcept;
	~AnyOfMatcher() override = default;
	AnyOfMatcher& operator=(const AnyOfMatcher&) = delete;
	AnyOfMatcher& operator=(AnyOfMatcher&&) noexcept;
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	std::tuple<MatcherTypes...> matchers_;
};

}
namespace qt_mocks
{

/**
 * @brief A matcher that ensures that all of the supplied matcher objects passed.
 *
 * @tparam ValueType This is the type the match objects will compare against.
 */
template<typename ValueType, typename ... MatcherTypes>
common_matchers::AllOfMatcher<ValueType, MatcherTypes...> allOf(MatcherTypes&& ... matchers);
/**
 * @brief A matcher that ensures that any of the supplied matcher objects passed.
 *
 * @tparam ValueType This is the type the match objects will compare against.
 */
template<typename ValueType, typename ... MatcherTypes>
common_matchers::AnyOfMatcher<ValueType, MatcherTypes...> anyOf(MatcherTypes&& ... matchers);
/**
 * @brief Creates a matcher that a value will be false against it.
 */
QT_MOCKS_EXPORT common_matchers::BooleanMatcher isFalse();
/**
 * @brief Creates a matcher that a value will be true against it.
 */
QT_MOCKS_EXPORT common_matchers::BooleanMatcher isTrue();
/**
 * @brief Creates a matcher that will expect that a value is equal to the matcher's expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expected The expected value of the test value.
 */
template<typename ValueType>
common_matchers::EqualMatcher<ValueType> equals(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is not equal to the matcher's expected
 * value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expected The expected value of the test value.
 */
template<typename ValueType>
common_matchers::NotEqualMatcher<ValueType> notEquals(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is less-than the matcher's expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expected The expected value of the test value.
 */
template<typename ValueType>
common_matchers::LessThanMatcher<ValueType> lessThan(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is less-than or equal to the matcher's
 * expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expected The expected value of the test value.
 */
template<typename ValueType>
common_matchers::LessThanEqualMatcher<ValueType> lessThanEqual(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is greater-than the matcher's expected
 * value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expected The expected value of the test value.
 */
template<typename ValueType>
common_matchers::GreaterThanMatcher<ValueType> greaterThan(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is greater-than or equal to the
 * matcher's expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expected The expected value of the test value.
 */
template<typename ValueType>
common_matchers::GreaterThanEqualMatcher<ValueType> greaterThanEqual(
	const ValueType& expected
);
/**
 * @brief Creates a matcher that will expect that a value is between two values, exclusive, to the
 * test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than this value.
 * @param endValue The expected value must be less than this value.
 */
template<typename ValueType>
common_matchers::BetweenExclusiveMatcher<ValueType> between(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will expect that a value is between two values,
 * beginning inclusive and ending exclusive, to the test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than or equal to this value.
 * @param endValue The expected value must be less than this value.
 */
template<typename ValueType>
common_matchers::BetweenBeginInclusiveMatcher<ValueType> betweenBeginInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will expect that a value is between two values,
 * beginning exclusive and ending inclusive, to the test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than this value.
 * @param endValue The expected value must be less than or equal to this value.
 */
template<typename ValueType>
common_matchers::BetweenEndInclusiveMatcher<ValueType> betweenEndInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will expect that a value is between two values,
 * beginning and ending inclusive, to the test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than or equal to this value.
 * @param endValue The expected value must be less than or equal to this value.
 */
template<typename ValueType>
common_matchers::BetweenInclusiveMatcher<ValueType> betweenInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will negate the result of another matcher.
 *
 * @tparam MatcherType The type of the matcher to be negated.
 *
 * @param matcher The matcher whose comparison result should be negated.
 */
template<typename MatcherType>
common_matchers::NegationMatcher<typename MatcherType::MatcherValueType> not_(
	MatcherType&& matcher
);
/**
 * @brief Creates a matcher that will expect that an object's member variable meets the conditions
 * of a matcher.
 *
 * @tparam ClassType The type of the class the member points to.
 * @tparam MemberVariableType The type of the member variable.
 * @tparam MatcherType The type of the matcher to be negated.
 *
 * @param memberVariable A pointer to the the member variable contained within the object to use
 * when comparing the actual value, the member variable, to the matcher's expected value.
 * @param matcher The matcher that tests the value of the object's member variable.
 */
template<typename ClassType, typename MemberVariableType, typename MatcherType>
common_matchers::MemberVariableMatcher<ClassType, MemberVariableType, MatcherType> field(
	MemberVariableType ClassType::* const memberVariable,
	const MatcherType& matcher
);
/**
 * @brief Creates a matcher that will expect that an object's no-argument member function meets the
 * conditions of a matcher. This function is most often the accessor to an object's private or
 * protected member variable.
 *
 * @tparam ClassType The type of the class the member points to.
 * @tparam MemberReturnType The return type of the member function.
 * @tparam MatcherType The type of the matcher to be negated.
 *
 * @param mumberFunction A pointer to the no-argument member function within the object to use to
 * obtain the actual value to compare against the matcher's expected value.
 * @param matcher The matcher that tests the value of the object's member function.
 */
template<typename ClassType, typename MemberReturnType, typename MatcherType>
common_matchers::MemberFunctionMatcher<ClassType, MemberReturnType, MatcherType> accessor(
	MemberReturnType (ClassType::* memberFunction)() const,
	MatcherType matcher
);
/**
 * @brief Creates a matcher for a container that is expected to be empty.
 *
 * @tparam ContainerType The type of the matcher to be negated.
 */
template<typename ContainerType>
common_matchers::IsEmptyMatcher<ContainerType> isEmpty();

}
namespace qt_mocks::internal
{

template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), bool> compareAll(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), bool> compareAll(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), std::string> descriptionAll(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), std::string> descriptionAll(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), bool> compareAny(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), bool> compareAny(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), std::string> descriptionAny(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), std::string> descriptionAny(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);

}

#include "Matchers.tpp"

#endif
