/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_MATCHERS_HPP_
#error Should not be included directly.
#endif

template<typename ValueType>
qt_mocks::common_matchers::ExpectedValue<ValueType>::ExpectedValue(ValueType value) :
	value_{std::move(value)}
{}

template<typename ValueType>
const ValueType& qt_mocks::common_matchers::ExpectedValue<ValueType>::value() const
{
	return this->value_;
}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(
	const ExpectedValue<ValueType>& value
) :
	qt_mocks::common_matchers::Matcher<ValueType>::Matcher{
		std::make_unique<ExpectedValue<ValueType>>(value)
	}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(
	std::unique_ptr<ExpectedValue<ValueType>> expected
) :
	expected_{std::move(expected)}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(const Matcher& other) :
	expected_{
		other.expected_ == nullptr
			? nullptr
			: std::make_unique<ExpectedValue<ValueType>>(other.expected_->value())
	}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(Matcher&& other) noexcept :
	expected_{std::move(other.expected_)}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>&
		qt_mocks::common_matchers::Matcher<ValueType>::operator=(Matcher&& other) noexcept
{
	if (this != &other)
	{
		this->expected_ = std::move(other.expected_);
	}
	return *this;
}

template<typename ValueType>
bool qt_mocks::common_matchers::EqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual == this->expected_->value();
}


template<typename ValueType>
std::string qt_mocks::common_matchers::EqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be equal to the expected value, "
			<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::NotEqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual != this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::NotEqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should not be equal to the expected value, "
		<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::LessThanMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual < this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::LessThanMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be less than the expected value, "
		<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::LessThanEqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual <= this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::LessThanEqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be less than or equal to the expected "
		"value, " << this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::GreaterThanMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual > this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::GreaterThanMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be greater than to the expected value, "
		<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual >= this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be greater than or equal to the expected "
		"value, " << this->expected_->value() << '.';
	return oss.str();
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
qt_mocks::common_matchers::MemberVariableMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::MemberVariableMatcher(
	MemberVariableType ClassType::* const member,
	const MatcherType& matcher
) :
	qt_mocks::common_matchers::Matcher<ClassType>{nullptr},
	member_{member},
	matcher_{matcher}
{}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
bool qt_mocks::common_matchers::MemberVariableMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::compare(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return this->matcher_.compare(actual.*this->member_);
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
std::string qt_mocks::common_matchers::MemberVariableMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::description(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return "The member variable with " + this->matcher_.description(actual.*this->member_);
}

template<typename ClassType, typename MemberFunctionType, typename MatcherType>
qt_mocks::common_matchers::MemberFunctionMatcher<
	ClassType,
	MemberFunctionType,
	MatcherType
>::MemberFunctionMatcher(
	MemberFunctionType (ClassType::* member)() const,
	const MatcherType& matcher
) :
	qt_mocks::common_matchers::Matcher<ClassType>{nullptr},
	member_{member},
	matcher_{matcher}
{}

template<typename ClassType, typename MemberFunctionType, typename MatcherType>
bool qt_mocks::common_matchers::MemberFunctionMatcher<
	ClassType,
	MemberFunctionType,
	MatcherType
>::compare(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return this->matcher_.compare((actual.*this->member_)());
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
std::string qt_mocks::common_matchers::MemberFunctionMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::description(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return "The member accessor should return with "
		+ this->matcher_.description((actual.*this->member_)());
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType>::BetweenExclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual > this->expected_->value()
		&& actual < this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
			<< " and " << this->end_ << ", exclusive.";
	return oss.str();
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType>::BetweenBeginInclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual >= this->expected_->value()
		&& actual < this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
			<< " and " << this->end_ << ", beginning inclusive and ending exclusive.";
	return oss.str();
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType>::BetweenEndInclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual > this->expected_->value()
		&& actual <= this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
		<< " and " << this->end_ << ", beginning exclusive and ending inclusive.";
	return oss.str();
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType>::BetweenInclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual >= this->expected_->value()
		&& actual <= this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
		<< " and " << this->end_ << ", beginning and ending inclusive.";
	return oss.str();
}

template<typename ValueType>
template<typename MatcherType>
qt_mocks::common_matchers::NegationMatcher<ValueType>::NegationMatcher(MatcherType&& matcher) :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matcher_{std::make_unique<MatcherType>(std::move(matcher))}
{}

template<typename ValueType>
bool qt_mocks::common_matchers::NegationMatcher<ValueType>::compare(const ValueType& actual) const
{
	return ! this->matcher_->compare(actual);
}

template<typename ValueType>
std::string qt_mocks::common_matchers::NegationMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	std::ostringstream oss;
	oss << "Expected the negation of the following: " << this->matcher_->description(actual);
	return oss.str();
}

template<typename Container>
qt_mocks::common_matchers::IsEmptyMatcher<Container>::IsEmptyMatcher() :
	qt_mocks::common_matchers::Matcher<Container>{nullptr}
{}

template<typename Container>
bool qt_mocks::common_matchers::IsEmptyMatcher<Container>::compare(const Container& actual) const
{
	return actual.empty();
}

template<typename Container>
std::string qt_mocks::common_matchers::IsEmptyMatcher<Container>::description(
	const Container& actual
) const
{
	std::ostringstream oss;
	oss << "The actual container is not empty. Contains: ";

	bool isFirst{true};
	for (const auto& value : actual)
	{
		oss << (isFirst ? "" : ", ") << value;
		isFirst = false;
	}
	oss << ')';
	return oss.str();
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::AllOfMatcher(
	MatcherTypes&& ... matchers
) :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::make_tuple(std::move(matchers)...)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::AllOfMatcher::AllOfMatcher(
	AllOfMatcher&& other
) noexcept :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::move(other.matchers_)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>&
		qt_mocks::common_matchers::AllOfMatcher<
			ValueType,
			MatcherTypes...
		>::AllOfMatcher::operator=(
	qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>&& other
) noexcept
{
	if (this != &other)
		this->matchers_ = std::move(other.matchers_);
	return *this;
}

template<typename ValueType, typename ... MatcherTypes>
bool qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::compare(
	const ValueType& actual
) const
{
	return internal::compareAll(actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
std::string qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::description(
	const ValueType& actual
) const
{
	std::ostringstream oss;
	return internal::descriptionAll(oss, actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::AnyOfMatcher(
	MatcherTypes&& ... matchers
) :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::make_tuple(std::move(matchers)...)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::AnyOfMatcher::AnyOfMatcher(
	AnyOfMatcher&& other
) noexcept :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::move(other.matchers_)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>&
		qt_mocks::common_matchers::AnyOfMatcher<
			ValueType,
			MatcherTypes...
		>::AnyOfMatcher::operator=(
	qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>&& other
) noexcept
{
	if (this != &other)
		this->matchers_ = std::move(other.matchers_);
	return *this;
}

template<typename ValueType, typename ... MatcherTypes>
bool qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::compare(
	const ValueType& actual
) const
{
	return internal::compareAny(actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
std::string qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::description(
	const ValueType& actual
) const
{
	std::ostringstream oss;
	return internal::descriptionAny(oss, actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...> qt_mocks::allOf(
	MatcherTypes&& ... matchers
)
{
	return common_matchers::AllOfMatcher<ValueType, MatcherTypes...>{
		std::forward<MatcherTypes>(matchers)...
	};
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...> qt_mocks::anyOf(
	MatcherTypes&& ... matchers
)
{
	return common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>{
		std::forward<MatcherTypes>(matchers)...
	};
}

template<typename ValueType>
qt_mocks::common_matchers::EqualMatcher<ValueType> qt_mocks::equals(const ValueType& expected)
{
	return common_matchers::EqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::NotEqualMatcher<ValueType> qt_mocks::notEquals(const ValueType& expected)
{
	return common_matchers::NotEqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::LessThanMatcher<ValueType> qt_mocks::lessThan(const ValueType& expected)
{
	return common_matchers::LessThanMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::LessThanEqualMatcher<ValueType> qt_mocks::lessThanEqual(
	const ValueType& expected
)
{
	return common_matchers::LessThanEqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::GreaterThanMatcher<ValueType> qt_mocks::greaterThan(
	const ValueType& expected
)
{
	return common_matchers::GreaterThanMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType> qt_mocks::greaterThanEqual(
	const ValueType& expected
)
{
	return common_matchers::GreaterThanEqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType> qt_mocks::between(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenExclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType> qt_mocks::betweenBeginInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenBeginInclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType> qt_mocks::betweenEndInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenEndInclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType> qt_mocks::betweenInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenInclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename MatcherType>
qt_mocks::common_matchers::NegationMatcher<
	typename MatcherType::MatcherValueType
> qt_mocks::not_(
	MatcherType&& matcher
)
{
	return common_matchers::NegationMatcher<typename MatcherType::MatcherValueType>{
		std::move(matcher)
	};
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
qt_mocks::common_matchers::MemberVariableMatcher<ClassType, MemberVariableType, MatcherType>
		qt_mocks::field(
	MemberVariableType ClassType::* const memberVariable,
	const MatcherType& matcher
)
{
	static_assert(
		std::is_member_object_pointer_v<decltype(memberVariable)>,
		"The supplied memberVariable must be a pointer to a class member variable."
	);
	return common_matchers::MemberVariableMatcher<ClassType, MemberVariableType, MatcherType>{
		memberVariable,
		matcher
	};
}

template<typename ClassType, typename MemberReturnType, typename MatcherType>
qt_mocks::common_matchers::MemberFunctionMatcher<ClassType, MemberReturnType, MatcherType>
		qt_mocks::accessor(
	MemberReturnType (ClassType::* memberFunction)() const,
	MatcherType matcher
)
{
	static_assert(
		std::is_member_function_pointer_v<decltype(memberFunction)>,
		"The supplied memberFunction must be a pointer to a class member function."
	);
	return common_matchers::MemberFunctionMatcher<ClassType, MemberReturnType, decltype(matcher)>{
		memberFunction,
		matcher
	};
}

template<typename ContainerType>
qt_mocks::common_matchers::IsEmptyMatcher<ContainerType> qt_mocks::isEmpty()
{
	return common_matchers::IsEmptyMatcher<ContainerType>{};
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), bool>
		qt_mocks::internal::compareAll(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return true;
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), bool>
		qt_mocks::internal::compareAll(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
	if ( ! std::get<Index>(matchers).compare(actual))
		return false;
	return compareAll<ValueType, Index + 1>(actual, matchers);
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), std::string>
		qt_mocks::internal::descriptionAll(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return oss.str();
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), std::string>
		qt_mocks::internal::descriptionAll(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
	if constexpr (Index == 0)
		oss << "Expected all of the following to be true. ";
	const auto& matcher{std::get<Index>(matchers)};
	const bool matchFailed{ ! matcher.compare(actual)};
	oss << (matchFailed ? "FAILED: " : "") << matcher.description(actual);

	if constexpr (Index < sizeof...(MatcherTypes) - 1)
		oss << ' ';
	return descriptionAll<ValueType, Index + 1>(oss, actual, matchers);
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), bool>
		qt_mocks::internal::compareAny(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return false;
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), bool>
		qt_mocks::internal::compareAny(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
	if (std::get<Index>(matchers).compare(actual))
		return true;
	return compareAny<ValueType, Index + 1>(actual, matchers);
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<Index == sizeof...(MatcherTypes), std::string>
		qt_mocks::internal::descriptionAny(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return oss.str();
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
std::enable_if_t<(Index < sizeof...(MatcherTypes)), std::string>
		qt_mocks::internal::descriptionAny(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
	if constexpr (Index == 0)
		oss << "Expected any of the following to be true. ";
	const auto& matcher{std::get<Index>(matchers)};
	const bool matchFailed{ ! matcher.compare(actual)};
	oss << (matchFailed ? "FAILED: " : "") << matcher.description(actual);

	if constexpr  (Index < sizeof...(MatcherTypes) - 1)
		oss << ' ';
	return descriptionAny<ValueType, Index + 1>(oss, actual, matchers);
}
