/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_MOCK_HPP
#define QT_MOCKS_MOCK_HPP

#include <functional>
#include <tuple>
#include <type_traits>

/**
 * @brief Generates a mock of a class's virtual member function for testing. The mocked object can
 * be used in unit tests to simulate and verify interactions with the interfaces they mock. Any
 * member function's mock that does not have a custom mock function will invoke a default empty
 * function that does nothing.
 *
 * @param ReturnType The return type of the member function to mock.
 * @param memberFunctionName The name of the member function to mock.
 * @param ParameterTypes The member function's argument types, surrounded in round brackets.
 * @param specifiers The member function's specifiers, surrounded in round brackets. When a member
 * function does not have any specifiers, a set of empty round brackets must be supplied. The
 * override specifier is always implied.
 *
 * To use the macro, include the necessary headers and invoke the macro with the desired interface
 * as illustrated below.
 * @code
 * struct Foo // Class to be mocked
 * {
 *     virtual ~Foo() = default;
 *
 *      [[nodiscard]]
 *      virtual qint32 bar() = 0;
 *
 *      virtual void bar(const qreal&, const qreal&, const qreal&) const noexcept = 0;
 * };
 *
 * #include "Mock.hpp"
 *
 * struct FooMock : Foo // Mocking class
 * {
 *      [[nodiscard]]
 *      QMOCK(qint32, bar, (), ());
 *
 *      QMOCK(void, bar, (const qreal&, const qreal&, const qreal&), (const, noexcept));
 * };
 *
 * @endcode
 *
 * @note Additional things to consider:
 * - The generated mock class will have a member variable per mocked member function.
 * - Each mocked member function variable stores the function to invoke and the arguments it's to be
 * invoked with.
 *
 * @bug Any member function that has the same name, number of arguments, and specifiers results in a
 * compiler error. The workaround is to add an extra comma, ",", in one of the mock's member
 * function specifier list.
 */
#define QMOCK(ReturnType, memberFunctionName, ParameterTypes, specifiers)  \
	ReturnType memberFunctionName ParameterTypes _QT_MOCKS_SPACE_SEPARATE specifiers override  \
	{  \
		auto& mock{this->_QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers)}; \
		\
		if constexpr (std::is_void_v<ReturnType>) \
		{ \
			if (mock.function_ != nullptr)  \
				std::apply(mock.function_, mock.arguments_); \
		} \
		else \
			return qt_mocks::internals::invokeImplementation(mock);  \
	}  \
	\
	mutable qt_mocks::internals::Mock<ReturnType, _QT_MOCKS_SPLIT_PARAMETERS ParameterTypes> _QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, ParameterTypes, specifiers)
/**
 * @brief Associates a mock function with a specified member function of a mocked object. The macro
 * is used to bind a user-provided function to a member function of a supplied mocked object,
 * allowing the user to specify the behavior of the mocked function when it is called.
 *
 * @param var The mocked object containing the member function to bind the mock to. This mock object
 * should be defined using the `QMOCK` macro.
 * @param memberFunctionName The name of the member function within the mock object to bind.
 * @param parametersList The parameters list of the member function to bind. This should match
 * the signature used when the member function was originally mocked with the `QMOCK` macro.
 * @param specifiersList The specifiers list for the member function to assist in identifying the
 * correct mocked function to bind to. It should match the specifiers provided in the `QMOCK` macro.
 *
 * @code
 * FooMock mock;
 * QMOCK_BIND(mock, bar, (), (const)) =
 *		[] ()
 *		{
 *			return 42;
 *		};
 * @endcode
 *
 * @note Make sure the function signature, including parameter types and const/noexcept qualifiers,
 * matches exactly with how it was defined in `QMOCK`. The binding affects the mock object's
 * behavior only during the test phase.
 *
 * @bug Misalignment between the function signature in `QMOCK` and the binding expression in
 * `QMOCK_BIND` can lead to unexpected behavior or compiler errors.
 */
#define QMOCK_BIND(var, memberFunctionName, parametersList, specifiersList)  \
	(var)._QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, parametersList, specifiersList).function_

namespace qt_mocks::internals
{
template<typename ReturnType, typename ... ParameterTypes>
struct Mock
{
	ReturnType gives() const;

	std::function<ReturnType(ParameterTypes ...)> function_{};
	std::tuple<std::remove_const_t<std::remove_reference_t<ParameterTypes>> ...> arguments_{};
};

template<typename ReturnType>
struct Mock<ReturnType, void>
{
	ReturnType gives() const;

	std::function<ReturnType()> function_{};
	std::tuple<> arguments_{};
};

template<typename ReturnType, typename ... ParameterTypes>
ReturnType invokeImplementation(Mock<ReturnType, ParameterTypes ...>& mock);
}

#define _QT_MOCKS_SPLIT_ARGS(...) __VA_ARGS__
#define _QT_MOCKS_IGNORE_1(_1, arg1, ...) arg1
#define _QT_MOCKS_IGNORE_2(...) _QT_MOCKS_IGNORE_1(__VA_ARGS__)

#define _QT_MOCKS_IS_VOID_PARAMETER() ignore, void
#define _QT_MOCKS_SPLIT_PARAMETERS_1(arg1) _QT_MOCKS_IGNORE_2(_QT_MOCKS_IS_VOID_PARAMETER arg1 (), arg1)
#define _QT_MOCKS_SPLIT_PARAMETERS_2(arg1, arg2) arg1, arg2
#define _QT_MOCKS_SPLIT_PARAMETERS_3(arg1, arg2, arg3) arg1, arg2, arg3
#define _QT_MOCKS_SPLIT_PARAMETERS_4(arg1, arg2, arg3, arg4) arg1, arg2, arg3, arg4
#define _QT_MOCKS_SPLIT_PARAMETERS_5(arg1, arg2, arg3, arg4, arg5) arg1, arg2, arg3, arg4, arg5
#define _QT_MOCKS_SPLIT_PARAMETERS_6(arg1, arg2, arg3, arg4, arg5, arg6) arg1, arg2, arg3, arg4, arg5, arg6
#define _QT_MOCKS_SPLIT_PARAMETERS_7(arg1, arg2, arg3, arg4, arg5, arg6, arg7) arg1, arg2, arg3, arg4, arg5, arg6, arg7
#define _QT_MOCKS_SPLIT_PARAMETERS_8(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8) arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8
#define _QT_MOCKS_SPLIT_PARAMETERS_9(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9) arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9
#define _QT_MOCKS_SPLIT_PARAMETERS_10(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10) arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10
#define _QT_MOCKS_SPLIT_PARAMETERS_CHOOSER(  \
		_1,  \
		_2,  \
		_3,  \
		_4,  \
		_5,  \
		_6,  \
		_7,  \
		_8,  \
		_9,  \
		_10,  \
		CHOSEN_ARITY,  \
		...  \
	) CHOSEN_ARITY
#define _QT_MOCKS_SPLIT_PARAMETERS(...)  \
	_QT_MOCKS_SPLIT_PARAMETERS_CHOOSER(__VA_ARGS__,  \
		_QT_MOCKS_SPLIT_PARAMETERS_10,  \
		_QT_MOCKS_SPLIT_PARAMETERS_9,  \
		_QT_MOCKS_SPLIT_PARAMETERS_8,  \
		_QT_MOCKS_SPLIT_PARAMETERS_7,  \
		_QT_MOCKS_SPLIT_PARAMETERS_6,  \
		_QT_MOCKS_SPLIT_PARAMETERS_5,  \
		_QT_MOCKS_SPLIT_PARAMETERS_4,  \
		_QT_MOCKS_SPLIT_PARAMETERS_3,  \
		_QT_MOCKS_SPLIT_PARAMETERS_2,  \
		_QT_MOCKS_SPLIT_PARAMETERS_1  \
	)(__VA_ARGS__)

#define _QT_MOCKS_IS_NO_PARAMETERS() ignore, 0
#define _QT_MOCKS_CALC_ARITY_1(arg1) _QT_MOCKS_IGNORE_2(_QT_MOCKS_IS_NO_PARAMETERS arg1 (), 1)
#define _QT_MOCKS_CALC_ARITY_2(arg1, arg2) 2
#define _QT_MOCKS_CALC_ARITY_3(arg1, arg2, arg3) 3
#define _QT_MOCKS_CALC_ARITY_4(arg1, arg2, arg3, arg4) 4
#define _QT_MOCKS_CALC_ARITY_5(arg1, arg2, arg3, arg4, arg5) 5
#define _QT_MOCKS_CALC_ARITY_6(arg1, arg2, arg3, arg4, arg5, arg6) 6
#define _QT_MOCKS_CALC_ARITY_7(arg1, arg2, arg3, arg4, arg5, arg6, arg7) 7
#define _QT_MOCKS_CALC_ARITY_8(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8) 8
#define _QT_MOCKS_CALC_ARITY_9(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9) 9
#define _QT_MOCKS_CALC_ARITY_10(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10) 10
#define _QT_MOCKS_CALC_ARITY_CHOOSER(  \
		_1,  \
		_2,  \
		_3,  \
		_4,  \
		_5,  \
		_6,  \
		_7,  \
		_8,  \
		_9,  \
		_10,  \
		CHOSEN_ARITY,  \
		...  \
	) CHOSEN_ARITY
#define _QT_MOCKS_CALC_ARITY(...)  \
	_QT_MOCKS_CALC_ARITY_CHOOSER(__VA_ARGS__,  \
		_QT_MOCKS_CALC_ARITY_10,  \
		_QT_MOCKS_CALC_ARITY_9,  \
		_QT_MOCKS_CALC_ARITY_8,  \
		_QT_MOCKS_CALC_ARITY_7,  \
		_QT_MOCKS_CALC_ARITY_6,  \
		_QT_MOCKS_CALC_ARITY_5,  \
		_QT_MOCKS_CALC_ARITY_4,  \
		_QT_MOCKS_CALC_ARITY_3,  \
		_QT_MOCKS_CALC_ARITY_2,  \
		_QT_MOCKS_CALC_ARITY_1  \
	)(__VA_ARGS__)

#define _QT_MOCKS_MOCK_VAR_NAME_ARITY_CREATOR(funcName, parameterArity, specifiersList)  \
		_## funcName ## _parameters_ ## parameterArity ## _specifers_ ## specifiersList ##_
#define _QT_MOCKS_MOCK_VAR_NAME_ARITY(funcName, parameterArity, specifiersList)  \
		_QT_MOCKS_MOCK_VAR_NAME_ARITY_CREATOR(funcName, parameterArity, specifiersList)
#define _QT_MOCKS_MOCK_VAR_NAME(memberFunctionName, parametersList, specifiersList) \
	_QT_MOCKS_MOCK_VAR_NAME_ARITY(  \
		memberFunctionName,  \
		_QT_MOCKS_CALC_ARITY(_QT_MOCKS_SPLIT_ARGS parametersList),  \
		_QT_MOCKS_SPECIFIER_SEPARATE specifiersList  \
	)

#define _QT_MOCKS_IS_NO_SPECIFIER() ignore, none
#define _QT_MOCKS_SPECIFIER_SEPARATE_1(arg1)   \
	_QT_MOCKS_IGNORE_2(  \
		_QT_MOCKS_IS_NO_SPECIFIER arg1 (),  \
		arg1  \
	)
#define _QT_MOCKS_SPECIFIER_SEPARATE_2(arg1, arg2) arg1 ##_## arg2
#define _QT_MOCKS_SPECIFIER_SEPARATE_3(arg1, arg2, arg3) arg1 ##_## arg2 ##_## arg3
#define _QT_MOCKS_SPECIFIER_SEPARATE_CHOOSER(_1, _2, _3, CHOSEN_ARITY, ...) CHOSEN_ARITY
#define _QT_MOCKS_SPECIFIER_SEPARATE(...)  \
	_QT_MOCKS_SPECIFIER_SEPARATE_CHOOSER(__VA_ARGS__,  \
		_QT_MOCKS_SPECIFIER_SEPARATE_3,  \
		_QT_MOCKS_SPECIFIER_SEPARATE_2,  \
		_QT_MOCKS_SPECIFIER_SEPARATE_1  \
	)(__VA_ARGS__)

#define _QT_MOCKS_IS_NO_SPACE() ignore,
#define _QT_MOCKS_SPACE_SEPARATE_1(arg1) _QT_MOCKS_IGNORE_2(_QT_MOCKS_IS_NO_SPACE arg1 (), arg1)
#define _QT_MOCKS_SPACE_SEPARATE_2(arg1, arg2) arg1 arg2
#define _QT_MOCKS_SPACE_SEPARATE_3(arg1, arg2, arg3) arg1 arg2 arg3
#define _QT_MOCKS_SPACE_SEPARATE_CHOOSER(_1, _2, _3, CHOSEN_ARITY, ...) CHOSEN_ARITY
#define _QT_MOCKS_SPACE_SEPARATE(...)  \
	_QT_MOCKS_SPACE_SEPARATE_CHOOSER(__VA_ARGS__,  \
		_QT_MOCKS_SPACE_SEPARATE_3,  \
		_QT_MOCKS_SPACE_SEPARATE_2,  \
		_QT_MOCKS_SPACE_SEPARATE_1  \
	)(__VA_ARGS__)

#include "Mock.tpp"

#endif
