/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_OUTPUT_HPP_
#define QT_MOCKS_OUTPUT_HPP_

#include <QList>
#include <QMap>
#include <QSet>
#include <QVector>

#include <cstdint>
#include <iomanip>
#include <map>
#include <ostream>
#include <set>
#include <type_traits>
#include <vector>

#include <qt-mocks/qt-mocks_export.h>


class QString;

namespace qt_mocks::internal
{

struct QT_MOCKS_EXPORT OstreamRef
{
	OstreamRef();
	// cppcheck-suppress noExplicitConstructor
	// Justification: Meant as a pass through type.


	OstreamRef(std::ostream& o);

	std::ostream& stream_;
};

template<typename, typename = void>
struct HasStreamInsertion : std::false_type
{};

template<typename T>
struct HasStreamInsertion<T, decltype(operator<<(OstreamRef{}, std::declval<T>()))> : std::true_type
{};

template<typename IteratorType>
void outputStream(
	std::ostream& stream,
	const IteratorType& begin,
	const IteratorType& end
);

}
namespace qt_mocks
{

template<typename KeyType, typename ElementType>
std::ostream& operator<<(std::ostream& stream, const std::map<KeyType, ElementType>& container);
template<typename FirstValueType, typename SecondValueType>
std::ostream& operator<<(std::ostream& stream, const std::pair<FirstValueType, SecondValueType>& p);
template<typename ElementType>
std::ostream& operator<<(std::ostream& stream, const std::set<ElementType>& container);
template<typename ElementType>
std::ostream& operator<<(std::ostream& stream, const std::vector<ElementType>& container);
QT_MOCKS_EXPORT std::ostream& operator<<(std::ostream& stream, const QString& str);
template<typename KeyType, typename ElementType>
std::ostream& operator<<(std::ostream& stream, const ::QMap<KeyType, ElementType>& container);
template<typename ElementType>
std::ostream& operator<<(std::ostream& stream, const ::QSet<ElementType>& container);
template<typename ElementType>
std::ostream& operator<<(std::ostream& stream, const ::QList<ElementType>& container);
template<typename MissingStreamInsertionType>
std::enable_if_t<
	 ! internal::HasStreamInsertion<MissingStreamInsertionType>::value,
	std::ostream&
> operator<<(internal::OstreamRef o, const MissingStreamInsertionType& object)
	// Justification: Visual Studio fails to locate the definition.
{
	const auto bytes{
		static_cast<const unsigned char *>(reinterpret_cast<const void*>(&object))
	};
	const auto size{sizeof object};
	const auto lastByte{size - 1};
	o.stream_ << "Raw Bytes [" << std::setw(2) << std::setfill('0') << std::hex;

	for(std::size_t byte{0} ; byte < size ; ++byte)
		o.stream_ << "0x" << static_cast<std::uint16_t>(bytes[byte])
			<< (byte == lastByte ? "" : " ");
	o.stream_ << ']';
	// cppcheck-suppress returnReference
	// Justification: Object is a reference to a stream.
	return o.stream_;
}

}


#include "Output.tpp"

#endif
